const { getClient } = require('../matrix-bot');

const { matrixUserToAuthorAddress } = require('../util/db');
const write = require('./write');
const { dao } = require('../util/contracts');

const {
  ETH_NETWORK,
} = process.env;

const addPostWithRetry = async (authors, hash, citations, retryDelay = 5000) => {
  try {
    await dao.addPost(authors, hash, citations);
  } catch (e) {
    if (e.code === 'REPLACEMENT_UNDERPRICED') {
      console.log('retry delay (sec):', retryDelay / 1000);
      await Promise.delay(retryDelay);
      return addPostWithRetry(authors, hash, citations, retryDelay * 2);
    } if (e.reason === 'A post with this contentId already exists') {
      return { alreadyAdded: true };
    }
    throw e;
  }
  return { alreadyAdded: false };
};

module.exports = async (req, res) => {
  const {
    body: {
      eventUri,
    },
  } = req;

  if (!eventUri) {
    res.status(400).end();
    return;
  }
  console.log(`importFromMatrix: event ${eventUri}`);

  // URI format:
  // https://matrix.to/#/${roomId}/${eventId}?via=
  const uriRegex = /#\/(![A-Za-z0-9:._-]+)\/(\$[A-Za-z0-9._-]+)(\?.*)$/;
  const [, roomId, eventId] = uriRegex.exec(new URL(eventUri).hash);
  console.log('roomId', roomId);
  console.log('eventId', eventId);

  const client = getClient();
  const event = await client.getEvent(roomId, eventId);
  console.log('event', event);

  let authorAddress;
  try {
    authorAddress = await matrixUserToAuthorAddress.get(event.sender);
  } catch (e) {
    // Matrix user has not registered their author address
    res.send(`Author address not registered for matrix user ${event.sender}`);
    return;
  }

  // We want to add a post representing this matrix message.
  // We can't sign it on behalf of the author.
  // That means we need to support posts without signatures.
  const authors = [{ authorAddress, weightPPM: 1000000 }];
  // TODO: Take citations as input to this API call, referencing other posts or matrix events
  const citations = [];
  const content = `Matrix event URI: ${eventUri}`;
  const embeddedData = {
    roomId,
    eventId,
  };

  const { hash } = await write({
    authors, citations, content, embeddedData,
  });

  // Now we want to add a post on-chain
  const { alreadyAdded } = await addPostWithRetry(authors, hash, citations);

  if (alreadyAdded) {
    console.log(`Post already added for matrix event ${eventUri}`);
  } else {
    console.log(`Added post to blockchain for matrix event ${eventUri}`);
    // Send matrix event reply to the targeted event, notifying of this blockchain post
    await client.replyNotice(roomId, event, `Added to ${ETH_NETWORK} blockchain as post ${hash}`);
  }

  res.json({ postId: hash, alreadyAdded });
};
