const axios = require('axios');
const ethers = require('ethers');
const crypto = require('crypto');
const Promise = require('bluebird');

const { authorAddresses, authorPrivKeys } = require('../util/db');
const { dao } = require('../util/contracts');
const write = require('./write');

// Each post allocates 30% of its reputation to citations
const PPM_TO_CITATIONS = 300000;

const fetchWithRetry = async (url, retryDelay = 5000) => {
  let retry = false;
  let responseData;
  let requestError;
  const response = await axios.get(url, {
    headers: {
      'api-key': process.env.SEMANTIC_SCHOLAR_API_KEY,
    },
  }).catch(async (error) => {
    if (error.response?.status === 429) {
      // Rate limit
      retry = true;
      return;
    }
    // Some other error occurred
    requestError = error;
  });
  if (requestError) {
    throw requestError;
  } else if (retry) {
    console.log('retry delay (sec):', retryDelay / 1000);
    await Promise.delay(retryDelay);
    responseData = await fetchWithRetry(url, retryDelay * 2);
  } else {
    responseData = response.data;
  }
  return responseData;
};

const fetchPaperInfo = async (paperId) => {
  const url = `https://api.semanticscholar.org/graph/v1/paper/${paperId}?`
  + 'fields=title,url,authors,references.title,references.url,references.authors';
  return fetchWithRetry(url);
};

const fetchAuthorPapers = async (authorId) => {
  const url = `https://api.semanticscholar.org/graph/v1/author/${authorId}/papers?`
  + 'fields=title,url,authors,references.title,references.url,references.authors';
  const responseData = await fetchWithRetry(url);
  // TODO: pagination
  return responseData.data;
};

const getOrCreateAuthors = async (paper) => Promise.mapSeries(
  paper.authors.filter((x) => !!x.authorId),
  async ({ authorId }) => {
  // Check if we already have an account for each author
    let authorAddress;
    let authorPrivKey;
    try {
      authorAddress = await authorAddresses.get(authorId);
    } catch (e) {
    // Probably not found
    }
    if (authorAddress) {
    // This should always succeed, so we don't use try/catch here
      authorPrivKey = await authorPrivKeys.get(authorAddress);
    } else {
    // Generate and store a new account
      const id = crypto.randomBytes(32).toString('hex');
      authorPrivKey = `0x${id}`;
      const wallet = new ethers.Wallet(authorPrivKey);
      authorAddress = wallet.address;
      await authorAddresses.put(authorId, authorAddress);
      await authorPrivKeys.put(authorAddress, authorPrivKey);
    }
    return {
      authorAddress,
      authorPrivKey,
    };
  },
);

const generatePost = async (paper) => {
  const authorsInfo = await getOrCreateAuthors(paper);
  if (!authorsInfo.length) {
    throw new Error('Paper has no authors with id');
  }
  const firstAuthorWallet = new ethers.Wallet(authorsInfo[0].authorPrivKey);
  const eachAuthorWeightPercent = Math.floor(1000000 / authorsInfo.length);
  const authors = authorsInfo.map(({ authorAddress }) => ({
    weightPPM: eachAuthorWeightPercent,
    authorAddress,
  }));
    // Make sure author weights sum to 100%
  const totalAuthorsWeight = authors.reduce((t, { weightPPM }) => t + weightPPM, 0);
  authors[0].weightPPM += 1000000 - totalAuthorsWeight;

  const content = `Semantic Scholar paper ${paper.paperId}
${paper.title}
HREF ${paper.url}`;

  const embeddedData = {
    semanticScholarPaperId: paper.paperId,
  };
  let contentToSign = content;
  if (embeddedData && Object.entries(embeddedData).length) {
    contentToSign += `\n\n${JSON.stringify(embeddedData, null, 2)}`;
  }
  const signature = firstAuthorWallet.signMessageSync(contentToSign);
  return {
    authors, content, signature, embeddedData,
  };
};

const addPostWithRetry = async (authors, hash, citations, retryDelay = 5000) => {
  try {
    await dao.addPost(authors, hash, citations);
  } catch (e) {
    if (e.code === 'REPLACEMENT_UNDERPRICED') {
      console.log('retry delay (sec):', retryDelay / 1000);
      await Promise.delay(retryDelay);
      return addPostWithRetry(authors, hash, citations, retryDelay * 2);
    } if (e.reason === 'A post with this contentId already exists') {
      return { alreadyAdded: true };
    }
    throw e;
  }
  return { alreadyAdded: false };
};

const importPaper = async (paper) => {
  console.log('references count:', paper.references.length);
  const { paperId } = paper;
  const references = paper.references.filter((x) => !!x.paperId);
  const eachCitationWeightPercent = Math.floor(PPM_TO_CITATIONS / references.length);
  const citations = (await Promise.mapSeries(
    references,
    async (citedPaper) => {
      // We need to fetch this paper so we can generate the post we WOULD add to the forum.
      // That way, if we later add the cited paper to the blockchain it will have the correct hash.
      // The forum allows dangling citations to support this use case.
      try {
        const citedPost = await generatePost(citedPaper);
        return {
          weightPPM: eachCitationWeightPercent,
          targetPostId: citedPost.hash,
        };
      } catch (e) {
        return null;
      }
    },
  )).filter((x) => !!x);

  // Make sure citation weights sum to the designated total
  if (citations.length) {
    const totalCitationWeight = citations.reduce((t, { weightPPM }) => t + weightPPM, 0);
    citations[0].weightPPM += PPM_TO_CITATIONS - totalCitationWeight;
  }

  // Create a post for this paper
  const {
    authors, content, signature, embeddedData,
  } = await generatePost(paper);

  // Write the new post to our database
  const hash = await write({
    authors, content, signature, embeddedData, citations,
  });

  // Add the post to the forum (on-chain)
  const { alreadyAdded } = await addPostWithRetry(authors, hash, citations);
  if (alreadyAdded) {
    console.log(`Post already added for paper ${paperId}`);
  } else {
    console.log(`Added post to blockchain for paper ${paperId}`);
  }

  return { paperId, postId: hash, alreadyAdded };
};

module.exports = async (req, res) => {
  const {
    body: {
      paperId,
      authorId,
    },
  } = req;

  if (paperId) {
    console.log(`importFromSS: paper ${paperId}`);
    const paper = await fetchPaperInfo(paperId);
    const result = await importPaper(paper);
    res.json(result);
  } else if (authorId) {
    console.log(`importFromSS: author ${authorId}`);
    const papers = await fetchAuthorPapers(authorId);
    console.log('papers count:', papers.length);
    const result = await Promise.mapSeries(papers, importPaper);
    if (result.length) {
      console.log(`Added posts for ${result.length} papers by author ${authorId}`);
    }
    res.json(result);
  } else {
    res.status(400).end();
  }

  // TODO: Send matrix room event on SS import
};
