import axios from 'axios';
import { Buffer } from 'buffer';
import { recoverPersonalSignature } from '@metamask/eth-sig-util';
import objectHash from 'object-hash/dist/object_hash';

// Make Buffer available to recoverPersonalSignature
window.Buffer = Buffer;

class Post {
  constructor({
    authors, content, signature, hash, embeddedData, citations,
  }) {
    this.authors = authors;
    this.content = content;
    this.signature = signature;
    this.hash = hash;
    this.embeddedData = embeddedData ?? {};
    this.citations = citations ?? [];
  }

  // Read from API
  static async read(hash) {
    const {
      data: {
        content, authors, signature, embeddedData, citations,
      },
    } = await axios.get(`/api/read/${hash}`);
    // Verify hash
    const derivedHash = objectHash({
      authors, content, signature, embeddedData,
    });
    if (hash !== derivedHash) {
      throw new Error('Hash mismatch');
    }
    if (signature) {
      // Verify signature
      let contentToVerify = content;
      if (embeddedData && Object.entries(embeddedData).length) {
        contentToVerify += `\n\n${JSON.stringify(embeddedData, null, 2)}`;
      }
      const recovered = recoverPersonalSignature({ data: contentToVerify, signature });
      const authorAddresses = authors.map((author) => author.authorAddress.toLowerCase());
      if (!authorAddresses.includes(recovered.toLowerCase())) {
        throw new Error('Signer is not among the authors');
      }
    }
    return new Post({
      content, authors, signature, hash, embeddedData, citations,
    });
  }

  // Include MetaMask signature
  async sign(web3Provider, account) {
    const author = this.authors?.find(({ authorAddress }) => authorAddress === account);
    if (!author) {
      throw new Error('Post must be signed by one of its authors');
    }
    let contentToSign = this.content;
    if (this.embeddedData && Object.entries(this.embeddedData).length) {
      contentToSign += `\n\n${JSON.stringify(this.embeddedData, null, 2)}`;
    }
    const msg = `0x${Buffer.from(contentToSign, 'utf8').toString('hex')}`;
    this.signature = await web3Provider.request({
      method: 'personal_sign',
      params: [msg, account],
    });
    return this;
  }

  // Write to API
  async write() {
    const data = {
      authors: this.authors,
      content: this.content,
      signature: this.signature,
      embeddedData: this.embeddedData,
      citations: this.citations,
    };
    const { data: hash } = await axios.post('/api/write', data);
    this.hash = hash;
  }

  // Upload hash to blockchain
  async publish(DAO, account) {
    await DAO.methods.addPost(this.authors, this.hash, this.citations ?? []).send({
      from: account,
      gas: 1000000,
    });
  }
}

export default Post;
